# Autogenerated by Astropy's setup.py on 2018-10-18 21:16:50 UTC
from __future__ import unicode_literals
import datetime

version = "3.0.5"
githash = "df4c75aa2013af9537412f2306c47e480ca71fb3"


major = 3
minor = 0
bugfix = 5

release = True
timestamp = datetime.datetime(2018, 10, 18, 21, 16, 50)
debug = False

astropy_helpers_version = "3.0.2"

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
